/******************************************************************************/
/* collcoll - cichlid remote data collection and collation                    */
/* copyright (c) 2004 jamie twycross, jpt@cs.nott.ac.uk                       */
/* released under the gnu general public license v2                           */
/******************************************************************************/

#include "collcoll.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <pthread.h>
#include <arpa/inet.h>

#define BACKLOG		10

static Collector *connected_collectors;

static float *data, *data_tmp;
static unsigned int data_len, data_size;
static pthread_mutex_t data_lock;

static int server_sd;
static pcap_handler packet_handler;

/* set up connections to collector servers */
void init_collectors(Collector *collectors)
{
	unsigned int i;
	
	for(i = 0; collectors[i].hostname; i++)
	{
		/* get ip address of collector */
		if(!(collectors[i].he = \
				gethostbyname(collectors[i].hostname)))
			goto init_error;
		/* create socket */
		collectors[i].sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		if(collectors[i].sd == -1)
			goto init_error;
		/* connect socket to remote collector */
		collectors[i].sa.sin_family = AF_INET;
		collectors[i].sa.sin_addr.s_addr = *((unsigned int *) \
				collectors[i].he->h_addr);
		collectors[i].sa.sin_port = htons(collectors[i].port);
		memset(&collectors[i].sa.sin_zero, 0, 8);
		if(connect(collectors[i].sd, (struct sockaddr *) \
				&(collectors[i].sa), sizeof(struct sockaddr)) \
				== -1)
			goto init_error;
	}

	/* set our own private copy for future use */
	connected_collectors = collectors;
	
	return;

	init_error:
		perror(COLLECTOR_ERROR);
		exit(EXIT_FAILURE);
}

/* close connections to collectors */
void term_collectors(void)
{
	unsigned int i;
	static unsigned char command = CMD_CLOSE;

	for(i = 0; connected_collectors[i].hostname; i++)
	{
		send(connected_collectors[i].sd, &command, 1, 0);
		close(connected_collectors[i].sd);
	}
}

/* query collector servers */
void poll_collectors(void)
{
	register unsigned int i, j;
	static unsigned char command = CMD_SEND_DATA;
	
	memset(data, 0, data_len);
	for(i = 0; connected_collectors[i].hostname; i++)
	{
		send(connected_collectors[i].sd, &command, 1, 0);
		if(recv(connected_collectors[i].sd, data_tmp, data_len, 0) \
				== data_len)
			for(j = 0; j < data_size; j++)
				data[j] += data_tmp[j];
	}
}

static void sigint_handler(int s)
{
	/* ctrl-c, so close server socket before exiting */
	close(server_sd);
	exit(EXIT_SUCCESS);
}

/* set up collector server to listen on given port */
void init_server(unsigned short int port)
{
	struct sockaddr_in sa;
	struct sigaction siga;
	
	/* create socket */
	server_sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if(server_sd == -1)
		goto init_error;
	
	/* connect socket to port */
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = INADDR_ANY;
	sa.sin_port = htons(port);
	memset(&sa.sin_zero, 0, 8);
	if(bind(server_sd, (struct sockaddr *) &sa, sizeof(struct sockaddr)) \
			== -1)
		goto init_error;
	if(listen(server_sd, BACKLOG) == -1)
		goto init_error;

	/* capture INT to exit cleanly */
	siga.sa_handler = sigint_handler;
	sigemptyset(&siga.sa_mask);
	siga.sa_flags = SA_RESTART;
	if(sigaction(SIGINT, &siga, NULL) == -1)
		goto init_error;
	
	return;

	init_error:
		perror(COLLECTOR_ERROR);
		exit(EXIT_FAILURE);	
}

/* send the data we've collected */
static void send_data(int sd)
{
	pthread_mutex_lock(&data_lock);
	/* send the data */
	send(sd, data, data_len, 0);
	/* reset data array */
	memset(data, 0, data_len);
	pthread_mutex_unlock(&data_lock);
}

/* handle connections to collector server */
static void handle_connection(int sd)
{
	int num_bytes;
	unsigned char command;
	short unsigned int flag = 1;
	
	while(flag)
	{
		/* wait for a command */
		num_bytes = recv(sd, &command, 1, 0);
		if(num_bytes != 1)
			continue;

		/* parse command */
		switch(command)
		{
			case CMD_CLOSE:  /* close the connection */
				flag = 0;
				continue;
				break;

			case CMD_SEND_DATA:  /* send our data */
				send_data(sd);
				break;

			default:  /* invalid command */
				continue;
				break;
		}
	}
}

/* listen for connections to collector server */
void start_server(void)
{
	int sd, sin_size;
	struct sockaddr_in sa;

	/* doesn't fork due to strange pthread interactions */
	while(1)
	{
		/* wait for connection */
		sin_size = sizeof(struct sockaddr_in);
		if((sd = accept(server_sd, (struct sockaddr *) &sa, \
					&sin_size)) == -1)
			goto server_error;
	
		/* handle connection */
		printf("opened connection from %s\n", inet_ntoa(sa.sin_addr));
		handle_connection(sd);
		
		close(sd);
		printf("closed connection from %s\n", inet_ntoa(sa.sin_addr));
	}
	
	server_error:
		perror(COLLECTOR_ERROR);
		exit(EXIT_FAILURE);
}

/* initialise storage for collected data */
void init_data(int len)
{
	data_size = len;
	data_len = len * sizeof(float);
	if(!(data = (float *) malloc(data_len)))
		goto init_error;
	if(!(data_tmp = (float *) malloc(data_len)))
		goto init_error;
	pthread_mutex_init(&data_lock, NULL);

	return;

	init_error:
		perror(COLLECTOR_ERROR);
		exit(EXIT_FAILURE);
}

/* add value to data array */
inline void add_data(unsigned int index, float value)
{
	pthread_mutex_lock(&data_lock);
	data[index] += value;
	pthread_mutex_unlock(&data_lock);
}

/* set value in data array (no mutex lock) */
inline void set_data_nolock(unsigned int index, float value)
{
	data[index] = value;
}

/* get value in data array (no mutex lock) */
inline float get_data_nolock(unsigned int index)
{
	return(data[index]);
}

/* basic packet sniffer */
static void packet_sniffer(char *device)
{
	pcap_t *handle;
	bpf_u_int32 net, mask;
	char errbuf[PCAP_ERRBUF_SIZE];

	if(pcap_lookupnet(device, &net, &mask, errbuf))
		if(errno != 99)  /* allow no IP on iterface */
			goto sniffer_error;
	if(!(handle = pcap_open_live(device, BUFSIZ, 0, -1, errbuf)))
		goto sniffer_error;
	pcap_loop(handle, -1, packet_handler, NULL);
		
	sniffer_error:
		perror(COLLECTOR_ERROR);
		exit(EXIT_FAILURE);
	
}

/* start packet sniffer */
void start_sniffer(pcap_handler handler, char *device)
{
	pthread_t data_writer;

	packet_handler = handler;
	pthread_create(&data_writer, NULL, (void *) &packet_sniffer, device);
}
